# The full functionality of MAGEMin is wrapped in ../gen/magemin_library.jl
# Yet, the routines here make it more convenient to use this from julia
import Base.show
using Base.Threads: @threads
using ProgressMeter

const VecOrMat = Union{Nothing, AbstractVector{Float64}, AbstractVector{<:AbstractVector{Float64}}}

export  retrieve_solution_phase_information, remove_phases,
        init_MAGEMin, finalize_MAGEMin, point_wise_minimization, convertBulk4MAGEMin, use_predefined_bulk_rock, define_bulk_rock, create_output,
        print_info, create_gmin_struct, pwm_init, pwm_run,
        single_point_minimization, multi_point_minimization, MAGEMin_Data, W_Data,
        Initialize_MAGEMin, Finalize_MAGEMin



"""
    Holds general information about solution phases
"""
mutable struct ss_infos
    ss_name :: String
    n_em    :: Int64
    n_xeos  :: Int64
    ss_em   :: Vector{String}
    ss_xeos :: Vector{String}
end

"""
    Holds general information about the database
"""
mutable struct db_infos
    db_name :: String
    db_info :: String
    data_ss :: Array{ss_infos}
    ss_name :: Array{String}
    data_pp :: Array{String}
end
        


"""
    Holds the MAGEMin databases & required structures for every thread
"""
mutable struct MAGEMin_Data{TypeGV, TypeZB, TypeDB, TypeSplxData}
    db          :: String
    gv          :: TypeGV
    z_b         :: TypeZB
    DB          :: TypeDB
    splx_data   :: TypeSplxData
end

"""
    Holds the overriding Ws parameters
0 = "mp", 1 = "mb", 2 = "ig", 3 = "igd", 4 = "um", 5 = "ige", 6 = "alk"
"""
mutable struct W_Data
    SS_id   :: Vector{Int64}
    SS_len  :: Vector{Int64}
    Ws      :: Vector{Matrix{Float64}}
end


"""
    Function to retrieve the general information of the databases
"""
function retrieve_solution_phase_information(dtb)

    db_inf  = db_infos[db_infos("mp", "Metapelite (White et al., 2014)", ss_infos[ss_infos("liq", 8, 7, ["none", "q4L", "abL", "kspL", "anL", "slL", "fo2L", "fa2L", "h2oL"], ["none", "q", "fsp", "na", "an", "ol", "x", "h2o"]), ss_infos("fsp", 3, 2, ["none", "ab", "an", "san"], ["none", "ca", "k"]), ss_infos("bi", 7, 6, ["none", "phl", "annm", "obi", "east", "tbi", "fbi", "mmbi"], ["none", "x", "m", "y", "f", "t", "Q"]), ss_infos("g", 5, 4, ["none", "py", "alm", "spss", "gr", "kho"], ["none", "x", "z", "m", "f"]), ss_infos("ep", 3, 2, ["none", "cz", "ep", "fep"], ["none", "f", "Q"]), ss_infos("ma", 6, 5, ["none", "mut", "celt", "fcelt", "pat", "ma", "fmu"], ["none", "x", "y", "f", "n", "c"]), ss_infos("mu", 6, 5, ["none", "mut", "cel", "fcel", "pat", "ma", "fmu"], ["none", "x", "y", "f", "n", "c"]), ss_infos("opx", 7, 6, ["none", "en", "fs", "fm", "mgts", "fopx", "mnopx", "odi"], ["none", "x", "m", "y", "f", "c", "Q"]), ss_infos("sa", 5, 4, ["none", "spr4", "spr5", "fspm", "spro", "ospr"], ["none", "x", "y", "f", "Q"]), ss_infos("cd", 4, 3, ["none", "crd", "fcrd", "hcrd", "mncd"], ["none", "x", "m", "h"]), ss_infos("st", 5, 4, ["none", "mstm", "fst", "mnstm", "msto", "mstt"], ["none", "x", "m", "f", "t"]), ss_infos("chl", 8, 7, ["none", "clin", "afchl", "ames", "daph", "ochl1", "ochl4", "f3clin", "mmchl"], ["none", "x", "y", "f", "m", "QAl", "Q1", "Q4"]), ss_infos("ctd", 4, 3, ["none", "mctd", "fctd", "mnct", "ctdo"], ["none", "x", "m", "f"]), ss_infos("sp", 4, 3, ["none", "herc", "sp", "mt", "usp"], ["none", "x", "y", "z"]), ss_infos("ilmm", 5, 4, ["none", "oilm", "dilm", "dhem", "geik", "pnt"], ["none", "i", "g", "m", "Q"])], ["liq", "fsp", "bi", "g", "ep", "ma", "mu", "opx", "sa", "cd", "st", "chl", "ctd", "sp", "ilmm"], ["q", "crst", "trd", "coe", "stv", "ky", "sill", "and", "ru", "sph", "O2", "H2O", "qfm", "qif", "nno", "hm", "cco"]), db_infos("mb", "Metabasite (Green et al., 2016)", ss_infos[ss_infos("sp", 4, 3, ["none", "herc", "sp", "mt", "usp"], ["none", "x", "y", "z"]), ss_infos("opx", 6, 5, ["none", "en", "fs", "fm", "mgts", "fopx", "odi"], ["none", "x", "y", "f", "c", "Q"]), ss_infos("fsp", 3, 2, ["none", "ab", "an", "san"], ["none", "ca", "k"]), ss_infos("liq", 9, 8, ["none", "q4L", "abL", "kspL", "wo1L", "sl1L", "fa2L", "fo2L", "watL", "anoL"], ["none", "q", "fsp", "na", "wo", "sil", "ol", "x", "yan"]), ss_infos("mu", 6, 5, ["none", "mu", "cel", "fcel", "pa", "mam", "fmu"], ["none", "x", "y", "f", "n", "c"]), ss_infos("ilm", 3, 2, ["none", "oilm", "dilm", "dhem"], ["none", "x", "Q"]), ss_infos("ol", 2, 1, ["none", "fo", "fa"], ["none", "x"]), ss_infos("hb", 11, 10, ["none", "tr", "tsm", "prgm", "glm", "cumm", "grnm", "a", "b", "mrb", "kprg", "tts"], ["none", "x", "y", "z", "a", "k", "c", "f", "t", "Q1", "Q2"]), ss_infos("ep", 3, 2, ["none", "cz", "ep", "fep"], ["none", "f", "Q"]), ss_infos("g", 4, 3, ["none", "py", "alm", "gr", "kho"], ["none", "x", "z", "f"]), ss_infos("chl", 7, 6, ["none", "clin", "afchl", "ames", "daph", "ochl1", "ochl4", "f3clin"], ["none", "x", "y", "f", "QAl", "Q1", "Q4"]), ss_infos("bi", 6, 5, ["none", "phl", "annm", "obi", "east", "tbi", "fbi"], ["none", "x", "y", "f", "t", "Q"]), ss_infos("dio", 7, 6, ["none", "jd", "di", "hed", "acmm", "om", "cfm", "jac"], ["none", "x", "j", "t", "c", "Qaf", "Qfm"]), ss_infos("abc", 2, 1, ["none", "abm", "anm"], ["none", "ca"])], ["sp", "opx", "fsp", "liq", "mu", "ilm", "ol", "hb", "ep", "g", "chl", "bi", "dio", "abc"], ["q", "crst", "trd", "coe", "law", "ky", "sill", "and", "ru", "sph", "sph", "ab", "H2O", "qfm", "qif", "nno", "hm", "cco"]), db_infos("ig", "Igneous (Holland et al., 2018)", ss_infos[ss_infos("spn", 8, 7, ["none", "nsp", "isp", "nhc", "ihc", "nmt", "imt", "pcr", "qndm"], ["none", "x", "y", "c", "t", "Q1", "Q2", "Q3"]), ss_infos("bi", 6, 5, ["none", "phl", "annm", "obi", "eas", "tbi", "fbi"], ["none", "x", "y", "f", "t", "Q"]), ss_infos("cd", 3, 2, ["none", "crd", "fcrd", "hcrd"], ["none", "x", "h"]), ss_infos("cpx", 10, 9, ["none", "di", "cfs", "cats", "crdi", "cess", "cbuf", "jd", "cen", "cfm", "kjd"], ["none", "x", "y", "o", "n", "Q", "f", "cr", "t", "k"]), ss_infos("ep", 3, 2, ["none", "cz", "ep", "fep"], ["none", "f", "Q"]), ss_infos("g", 6, 5, ["none", "py", "alm", "gr", "andr", "knom", "tig"], ["none", "x", "c", "f", "cr", "t"]), ss_infos("hb", 11, 10, ["none", "tr", "tsm", "prgm", "glm", "cumm", "grnm", "a", "b", "mrb", "kprg", "tts"], ["none", "x", "y", "z", "a", "k", "c", "f", "t", "Q1", "Q2"]), ss_infos("ilm", 3, 2, ["none", "oilm", "dilm", "dhem"], ["none", "x", "Q"]), ss_infos("liq", 12, 11, ["none", "q4L", "slL", "wo1L", "fo2L", "fa2L", "jdL", "hmL", "ekL", "tiL", "kjL", "ctL", "wat1L"], ["none", "wo", "sl", "fo", "fa", "jd", "hm", "ek", "ti", "kj", "yct", "h2o"]), ss_infos("ol", 4, 3, ["none", "mont", "fa", "fo", "cfm"], ["none", "x", "c", "Q"]), ss_infos("opx", 9, 8, ["none", "en", "fs", "fm", "odi", "mgts", "cren", "obuf", "mess", "ojd"], ["none", "x", "y", "c", "Q", "f", "t", "cr", "j"]), ss_infos("fsp", 3, 2, ["none", "ab", "an", "san"], ["none", "ca", "k"]), ss_infos("fl", 11, 10, ["none", "qfL", "slfL", "wofL", "fofL", "fafL", "jdfL", "hmfL", "ekfL", "tifL", "kjfL", "H2O"], ["none", "wo", "sl", "fo", "fa", "jd", "hm", "ek", "ti", "kj", "h2o"]), ss_infos("mu", 6, 5, ["none", "mu", "cel", "fcel", "pa", "mam", "fmu"], ["none", "x", "y", "f", "n", "c"]), ss_infos("fper", 2, 1, ["none", "per", "wu"], ["none", ""])], ["spn", "bi", "cd", "cpx", "ep", "g", "hb", "ilm", "liq", "ol", "opx", "fsp", "fl", "mu", "fper"], ["q", "crst", "trd", "coe", "stv", "ky", "sill", "and", "ru", "sph", "O2", "qfm", "mw", "qif", "nno", "hm", "cco"]), db_infos("um", "Ultramafic (Evans & Frost., 2021)", ss_infos[ss_infos("fl", 2, 1, ["none", "H2", "H2O"], ["none", "x"]), ss_infos("ol", 2, 1, ["none", "fo", "fa"], ["none", "x"]), ss_infos("br", 2, 1, ["none", "br", "fbr"], ["none", "x"]), ss_infos("ch", 2, 1, ["none", "chum", "chuf"], ["none", "x"]), ss_infos("atg", 5, 4, ["none", "atgf", "fatg", "atgo", "aatg", "oatg"], ["none", "x", "y", "f", "t"]), ss_infos("g", 2, 1, ["none", "py", "alm"], ["none", "x"]), ss_infos("ta", 6, 5, ["none", "ta", "fta", "tao", "tats", "ota", "tap"], ["none", "x", "y", "f", "v", "Q"]), ss_infos("chl", 7, 6, ["none", "clin", "afchl", "ames", "daph", "ochl1", "ochl4", "f3clin"], ["none", "x", "y", "f", "m", "t", "QA1"]), ss_infos("spi", 3, 2, ["none", "herc", "sp", "mt"], ["none", "x", "y"]), ss_infos("opx", 5, 4, ["none", "en", "fs", "fm", "mgts", "fopx"], ["none", "x", "y", "f", "Q"]), ss_infos("po", 2, 1, ["none", "trov", "trot"], ["none", "y"]), ss_infos("anth", 5, 4, ["none", "anth", "gedf", "fant", "a", "b"], ["none", "x", "y", "z", "a"])], ["fl", "ol", "br", "ch", "atg", "g", "ta", "chl", "spi", "opx", "po", "anth"], ["q", "crst", "trd", "coe", "stv", "ky", "sill", "and", "pyr", "O2", "qfm", "qif", "nno", "hm", "cco"])]
    dbs     = ["mp","mb","ig","um"]
    id      = findall(dbs .== dtb)[1]

    return db_inf[id]
end

"""
    Function to retrieve the list of indexes of the solution phases to be removed from the minimization
"""
function remove_phases(list,dtb)
    if ~isnothing(list)
        db_inf = retrieve_solution_phase_information(dtb);
        rm_list = zeros(Int64,0);
        for i in list
            idx = findall(db_inf.ss_name .== i)
            if ~isempty(idx)
                rm_list = vcat(rm_list,idx[1]);
            else
                print(" \"$i\" is not a proper solid solution phase name, and thus cannot be deactivated")
            end
        end

        if isempty(rm_list)
            rm_list = nothing
            print(" The list of phases to be removed appears to be empty...")
        end
    else
        rm_list = nothing
    end
    
    return rm_list;
end



"""
    Dat = Initialize_MAGEMin(db = "ig"; verbose::Union{Bool, Int64} = true)

Initializes MAGEMin on one or more threads, for the database `db`. You can suppress all output with `verbose=false`. `verbose=true` will give a brief summary of the result, whereas `verbose=1` will give more details about the computations.
"""
function Initialize_MAGEMin(db = "ig";  verbose     ::Union{Int64,Bool} = 0,
                                        limitCaOpx  ::Int64             = 0,
                                        CaOpxLim    ::Float64           = 0.0,
                                        mbCpx       ::Int64             = 1,
                                        buffer      ::String            = "NONE",
                                        solver      ::Int64             = 1         )

    gv, z_b, DB, splx_data = init_MAGEMin(db;   verbose     = verbose,
                                                mbCpx       = mbCpx,
                                                limitCaOpx  = limitCaOpx,
                                                CaOpxLim    = CaOpxLim,
                                                buffer      = buffer,
                                                solver      = solver    );

    nt              = Threads.nthreads()
    list_gv         = Vector{typeof(gv)}(undef, nt)
    list_z_b        = Vector{typeof(z_b)}(undef, nt)
    list_DB         = Vector{typeof(DB)}(undef, nt)
    list_splx_data  = Vector{typeof(splx_data)}(undef, nt)

    if isa(verbose,Bool)
        if verbose
            verbose=0
        else
            verbose=-1
        end
    end

    for id in 1:nt
        gv, z_b, DB, splx_data = init_MAGEMin(db;   verbose     = verbose,
                                                    mbCpx       = mbCpx,
                                                    limitCaOpx  = limitCaOpx,
                                                    CaOpxLim    = CaOpxLim,
                                                    buffer      = buffer,
                                                    solver      = solver    );

        list_gv[id]         = gv
        list_z_b[id]        = z_b
        list_DB[id]         = DB
        list_splx_data[id]  = splx_data
    end

    return MAGEMin_Data(db, list_gv, list_z_b, list_DB, list_splx_data)
end


"""
    Finalize_MAGEMin(dat::MAGEMin_Data)
Finalizes MAGEMin and clears variables
"""
function Finalize_MAGEMin(dat::MAGEMin_Data)
    for id in 1:Threads.nthreads()

        LibMAGEMin.FreeDatabases(dat.gv[id], dat.DB[id], dat.z_b[id])
        # splx_data needs to be freed
     end
     return nothing
end


"""
    gv, DB = init_MAGEMin(;EM_database=0)

Initializes MAGEMin (including setting global options) and loads the Database.
"""
function  init_MAGEMin( db          =  "ig";
                        verbose     =   0,
                        mbCpx       =   0,
                        limitCaOpx  =   0,
                        CaOpxLim    =   1.0,
                        buffer      =  "NONE",
                        solver      =   1           )

    z_b         = LibMAGEMin.bulk_infos()
    gv          = LibMAGEMin.global_variables()
    splx_data   = LibMAGEMin.simplex_data();
    DB          = LibMAGEMin.Database()
    gv          = LibMAGEMin.global_variable_alloc( pointer_from_objref(z_b))

    if db == "ig"
        gv.EM_database = 2
    elseif db == "mb"
        gv.EM_database = 1
    elseif db == "mp"
        gv.EM_database = 0
    elseif db == "igd"
        gv.EM_database = 3
    elseif db == "ige"
        gv.EM_database = 5
    elseif db == "alk"
        gv.EM_database = 6
    elseif db == "um"
        gv.EM_database = 4
    else
        print("Database not implemented...\n")
    end

    gv.verbose      = verbose
    gv.mbCpx        = mbCpx
    gv.limitCaOpx   = limitCaOpx
    gv.CaOpxLim     = CaOpxLim
    gv.solver       = solver
    gv.buffer       = pointer(buffer)

    gv              = LibMAGEMin.global_variable_init(gv, pointer_from_objref(z_b))
    DB              = LibMAGEMin.InitializeDatabases(gv, gv.EM_database)

    LibMAGEMin.init_simplex_A(      pointer_from_objref(splx_data), gv)
    LibMAGEMin.init_simplex_B_em(   pointer_from_objref(splx_data), gv)

    return gv, z_b, DB, splx_data
end

"""
    finalize_MAGEMin(gv,DB)
Cleans up the memory
"""
function  finalize_MAGEMin(gv,DB, z_b)
    LibMAGEMin.FreeDatabases(gv, DB, z_b)
    nothing
end


# wrapper for single point minimization
function single_point_minimization(     P           ::  T1,
                                        T           ::  T1,
                                        MAGEMin_db  ::  MAGEMin_Data;
                                        test        ::  Int64                           = 0, # if using a build-in test case,
                                        X           ::  VecOrMat                        = nothing,      
                                        B           ::  Union{Nothing, T1, Vector{T1}}  = nothing,
                                        scp         = 0,   
                                        rm_list     ::  Union{Nothing, Vector{Int64}}           = nothing,
                                        W           ::  Union{Nothing, W_Data}          = nothing,
                                        Xoxides     = Vector{String},
                                        sys_in      = "mol",
                                        progressbar = true        # show a progress bar or not?
                                        ) where {T1 <: Float64}

    P   = [P];
    T   = [T];
    if X isa AbstractVector{Float64}
        X = [X]
    end

    Out_PT     =   multi_point_minimization(    P,
                                                T,
                                                MAGEMin_db,
                                                test        =   test,
                                                X           =   X,
                                                B           =   B,
                                                scp         =   scp,
                                                rm_list     =   rm_list,
                                                W           =   W,
                                                Xoxides     =   Xoxides,
                                                sys_in      =   sys_in,
                                                progressbar =   progressbar);

    return Out_PT[1]
end


"""
Out_PT =multi_point_minimization(P::T2,T::T2,MAGEMin_db::MAGEMin_Data;test::Int64=0,X::Union{Nothing, AbstractVector{Float64}, AbstractVector{<:AbstractVector{Float64}}}=nothing,B::Union{Nothing, T1, Vector{T1}}=nothing,W::Union{Nothing, W_Data}=nothing,Xoxides=Vector{String},sys_in="mol",progressbar=true) where {T1 <: Float64, T2 <: AbstractVector{T1}}


Perform (parallel) MAGEMin calculations for a range of points as a function of pressure `P`, temperature `T` and/or composition `X`. The database `MAGEMin_db` must be initialised before calling the routine.
The bulk-rock composition can either be set to be one of the pre-defined build-in test cases, or can be specified specifically by passing `X`, `Xoxides` and `sys_in` (that specifies whether the input is in "mol" or "wt").

Below a few examples:

Example 1 - build-in test vs. pressure and temperature
===
```julia
julia> data = Initialize_MAGEMin("ig", verbose=false);
julia> n = 10
julia> P = rand(8:40.0,n)
julia> T = rand(800:1500.0,n)
julia> out = multi_point_minimization(P, T, data, test=0)
julia> Finalize_MAGEMin(data)
```

Example 2 - Specify constant bulk rock composition for all points:
===
```julia
julia> data = Initialize_MAGEMin("ig", verbose=false);
julia> n = 10
julia> P = fill(10.0,n)
julia> T = fill(1100.0,n)
julia> Xoxides = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "Fe2O3"; "K2O"; "Na2O"; "TiO2"; "Cr2O3"; "H2O"];
julia> X = [48.43; 15.19; 11.57; 10.13; 6.65; 1.64; 0.59; 1.87; 0.68; 0.0; 3.0];
julia> sys_in = "wt"
julia> out = multi_point_minimization(P, T, data, X=X, Xoxides=Xoxides, sys_in=sys_in)
julia> Finalize_MAGEMin(data)
```

Example 3 - Different bulk rock composition for different points
===
```julia
julia> data = Initialize_MAGEMin("ig", verbose=false);
julia> P = [10.0, 20.0]
julia> T = [1100.0, 1200]
julia> Xoxides = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "Fe2O3"; "K2O"; "Na2O"; "TiO2"; "Cr2O3"; "H2O"];
julia> X1 = [48.43; 15.19; 11.57; 10.13; 6.65; 1.64; 0.59; 1.87; 0.68; 0.0; 3.0];
julia> X2 = [49.43; 14.19; 11.57; 10.13; 6.65; 1.64; 0.59; 1.87; 0.68; 0.0; 3.0];
julia> X = [X1,X2]
julia> sys_in = "wt"
julia> out = multi_point_minimization(P, T, data, X=X, Xoxides=Xoxides, sys_in=sys_in)
julia> Finalize_MAGEMin(data)
```

Activating multithreading on julia
===

To take advantage of multithreading, you need to start julia from the terminal with:
```bash
\$ julia -t auto
```
which will automatically use all threads on your machine. Alternatively, use `julia -t 4` to start it on 4 threads.
If you are interested to see what you can do on your machine type:
```
julia> versioninfo()
```

"""
function multi_point_minimization(P           ::  T2,
                                  T           ::  T2,
                                  MAGEMin_db  ::  MAGEMin_Data;
                                  test        ::  Int64                           = 0, # if using a build-in test case,
                                  X           ::  VecOrMat                        = nothing,
                                  B           ::  Union{Nothing, T1, Vector{T1}}  = nothing,
                                  scp                                             = 0,
                                  rm_list     ::  Union{Nothing, Vector{Int64}}   = nothing,
                                  W           ::  Union{Nothing, W_Data}          = nothing,
                                  Xoxides     = Vector{String},
                                  sys_in      = "mol",
                                  progressbar = true        # show a progress bar or not?
                                  ) where {T1 <: Float64, T2 <: AbstractVector{Float64}}

    # Set the compositional info
    CompositionType::Int64 = 0;

    if isnothing(X)
        # Use one of the build-in tests
        # Create thread-local data
        for i in 1:Threads.nthreads()
            MAGEMin_db.gv[i] = use_predefined_bulk_rock(MAGEMin_db.gv[i], test, MAGEMin_db.db)
        end
        CompositionType = 0;    # build-in tests
    else
        if isa(X,Vector{Float64})
        # same bulk rock composition for the full diagram
        @assert length(X) == length(Xoxides)

            # Set the bulk rock composition for all points
            for i in 1:Threads.nthreads()
                MAGEMin_db.gv[i] = define_bulk_rock(MAGEMin_db.gv[i], X, Xoxides, sys_in, MAGEMin_db.db);
            end
            CompositionType = 1;    # specified bulk composition for all points
        else
            @assert length(X) == length(P)
            CompositionType = 2;    # different bulk rock composition for every point
        end
    end

    # initialize vectors
    Out_PT = Vector{gmin_struct{Float64, Int64}}(undef, length(P))

    # main loop
    if progressbar
        progr = Progress(length(P), desc="Computing $(length(P)) points...") # progress meter
    end
    @threads :static for i in eachindex(P)
        # Get thread-local buffers. As of Julia v1.9, a dynamic scheduling of
        # the threads is the default setting. To avoid task migration and the
        # resulting concurrency issues, we restrict the loop to static scheduling.
        id          = Threads.threadid()
        gv          = MAGEMin_db.gv[id]
        z_b         = MAGEMin_db.z_b[id]
        DB          = MAGEMin_db.DB[id]
        splx_data   = MAGEMin_db.splx_data[id]

        if CompositionType==2
            # different bulk-rock composition for every point - specify it here
            gv = define_bulk_rock(gv, X[i], Xoxides, sys_in, MAGEMin_db.db);
        end

        # compute a new point using a ccall
        if isnothing(B)
            out     = point_wise_minimization(P[i], T[i], gv, z_b, DB, splx_data; scp, rm_list)
        else
            out     = point_wise_minimization(P[i], T[i], gv, z_b, DB, splx_data; buffer_n = B[i], W = W, scp, rm_list)
        end
        Out_PT[i]   = deepcopy(out)

        if progressbar
            next!(progr)
        end
    end
    if progressbar
        finish!(progr)
    end

    return Out_PT
end

"""
bulk_rock = use_predefined_bulk_rock(gv, test=-1, db="ig")

Returns the pre-defined bulk rock composition of a given test
"""
function use_predefined_bulk_rock(gv, test=0, db="ig")

    if db == "ig"
        gv.test = test
        gv = LibMAGEMin.get_bulk_igneous(gv)
        LibMAGEMin.norm_array(gv.bulk_rock, gv.len_ox)
    elseif db == "igd"
        gv.test = test
        gv = LibMAGEMin.get_bulk_igneous(gv)
        LibMAGEMin.norm_array(gv.bulk_rock, gv.len_ox)
    elseif db == "ige"
        gv.test = test
        gv = LibMAGEMin.get_bulk_igneous(gv)
        LibMAGEMin.norm_array(gv.bulk_rock, gv.len_ox)
    elseif db == "mp"
        gv.test = test
        gv = LibMAGEMin.get_bulk_metapelite(gv)
        LibMAGEMin.norm_array(gv.bulk_rock, gv.len_ox)
    elseif db == "mb"
        gv.test = test
        gv = LibMAGEMin.get_bulk_metabasite(gv)
        LibMAGEMin.norm_array(gv.bulk_rock, gv.len_ox)
    elseif db == "um"
        gv.test = test
        gv = LibMAGEMin.get_bulk_ultramafic(gv)
        LibMAGEMin.norm_array(gv.bulk_rock, gv.len_ox)
    elseif db == "alk"
        gv.test = test
        gv = LibMAGEMin.get_bulk_igneous_alk(gv)
        LibMAGEMin.norm_array(gv.bulk_rock, gv.len_ox)
    else
        print("Database not implemented...\n")
    end

    return gv
end

"""
    data = use_predefined_bulk_rock(data::MAGEMin_Data, test=0)
Returns the pre-defined bulk rock composition of a given test
"""
function use_predefined_bulk_rock(data::MAGEMin_Data, test=0)
    nt = Threads.nthreads()
    for id in 1:nt
        data.gv[id] =  use_predefined_bulk_rock(data.gv[id], test, data.db)
    end
    return data
end

function define_bulk_rock(gv, bulk_in, bulk_in_ox, sys_in,db)

    bulk_rock, ox   = convertBulk4MAGEMin(bulk_in,bulk_in_ox,sys_in,db)     # conversion changes the system unit to mol
    unsafe_copyto!(gv.bulk_rock, pointer(bulk_rock) , gv.len_ox)            # copy the bulk-rock

    LibMAGEMin.norm_array(gv.bulk_rock, gv.len_ox)

    return gv
end


function normalize(vector::Vector{Float64})
    return vector ./ sum(vector)
end


"""
convertBulk4MAGEMin( bulk_in, bulk_in_ox, sys_in)

receives bulk-rock composition in [mol,wt] fraction and associated oxide list and sends back bulk-rock composition converted for MAGEMin use

"""
function convertBulk4MAGEMin(bulk_in::T1,bulk_in_ox::Vector{String},sys_in::String,db::String) where {T1 <: AbstractVector{Float64}}

	ref_ox          = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "Fe2O3"; "K2O"; "Na2O"; "TiO2"; "O"; "Cr2O3"; "MnO"; "H2O"; "S"];
	ref_MolarMass   = [60.08; 101.96; 56.08; 40.30; 71.85; 159.69; 94.2; 61.98; 79.88; 16.0; 151.99; 70.937; 18.015; 32.06];      #Molar mass of oxides

    if db == "ig"
	    MAGEMin_ox      = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "K2O"; "Na2O"; "TiO2"; "O"; "Cr2O3"; "H2O"];
    elseif db == "igd"
        MAGEMin_ox      = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "K2O"; "Na2O"; "TiO2"; "O"; "Cr2O3"; "H2O"];
    elseif db == "ige"
        MAGEMin_ox      = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "K2O"; "Na2O"; "TiO2"; "O"; "Cr2O3"; "H2O"];
    elseif db == "alk"
        MAGEMin_ox      = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "K2O"; "Na2O"; "TiO2"; "O"; "Cr2O3"; "H2O"];
    elseif db == "mb"
        MAGEMin_ox      = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "K2O"; "Na2O"; "TiO2"; "O"; "H2O"];
    elseif db == "um"
        MAGEMin_ox      = ["SiO2"; "Al2O3"; "MgO"; "FeO"; "O"; "H2O"; "S"];
    elseif db == "mp"
        MAGEMin_ox      = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "K2O"; "Na2O"; "TiO2"; "O"; "MnO"; "H2O"];
    else
        print("Database not implemented...\n")
    end

	MAGEMin_bulk    = zeros(length(MAGEMin_ox));
    bulk            = zeros(length(MAGEMin_ox));
    
	# convert to mol, if system unit = wt
	if sys_in == "wt"
		for i=1:length(bulk_in_ox)
            id = findall(ref_ox .== bulk_in_ox[i])[1];
			bulk[i] = bulk_in[i]/ref_MolarMass[id];
		end
    else
		for i=1:length(bulk_in_ox)
			bulk[i] = bulk_in[i];
		end
	end

	bulk = normalize(bulk);

	for i=1:length(MAGEMin_ox)
        id = findall(bulk_in_ox .== MAGEMin_ox[i]);
		if ~isempty(id)
			MAGEMin_bulk[i] = bulk[id[1]];
		end
	end
    idFe2O3 = findall(bulk_in_ox .== "Fe2O3");

    if ~isempty(idFe2O3)
        idFeO = findall(MAGEMin_ox .== "FeO");
        MAGEMin_bulk[idFeO[1]] += bulk[idFe2O3[1]]*2.0;

        idO = findall(MAGEMin_ox .== "O");
        MAGEMin_bulk[idO[1]] += bulk[idFe2O3[1]];
    end

    MAGEMin_bulk .= normalize(MAGEMin_bulk);

    # check which component can safely be put to 0.0
    d = []
    c = []
    if db == "ig" || db == "igd" || db == "ige" ||  db == "alk"
        c = findall(MAGEMin_ox .!= "Cr2O3" .&& MAGEMin_ox .!= "TiO2" .&& MAGEMin_ox .!= "O" .&& MAGEMin_ox .!= "H2O");
        d = findall(MAGEMin_ox .== "Cr2O3" .|| MAGEMin_ox .== "TiO2" .|| MAGEMin_ox .== "O");# .|| MAGEMin_ox .== "H2O");
    elseif db == "mb"               #for the metabasite database it is better to set a low value for H2O as dry system haw not been validated by Eleanor
        c = findall(MAGEMin_ox .!= "TiO2" .&& MAGEMin_ox .!= "O");
        d = findall(MAGEMin_ox .== "TiO2" .|| MAGEMin_ox .== "O");
    elseif db == "mp"
        c = findall(MAGEMin_ox .!= "TiO2" .&& MAGEMin_ox .!= "O" .&& MAGEMin_ox .!= "MnO" .&& MAGEMin_ox .!= "H2O");
        d = findall(MAGEMin_ox .== "TiO2" .|| MAGEMin_ox .== "O" .|| MAGEMin_ox .!= "MnO");
    else
        c = findall(MAGEMin_ox .!= "H2O");
    end

    id0 = findall(MAGEMin_bulk[c] .< 1e-4)
    if ~isempty(id0)
        MAGEMin_bulk[c[id0]] .= 1e-4;
    end

    id1 = findall(MAGEMin_bulk[d] .< 2e-5)
    if ~isempty(id1)
        MAGEMin_bulk[d[id1]] .= 0.0;
    end
    MAGEMin_bulk .= normalize(MAGEMin_bulk).*100.0

    return MAGEMin_bulk, MAGEMin_ox;
end


"""
    point_wise_minimization(P::Float64,T::Float64, gv, z_b, DB, splx_data, sys_in::String="mol")

Computes the stable assemblage at `P` [kbar], `T` [C] and for a given bulk rock composition


# Example 1

This is an example of how to use it for a predefined bulk rock composition:
```julia
julia> db          = "ig"  # database: ig, igneous (Holland et al., 2018); mp, metapelite (White et al 2014b)
julia> gv, z_b, DB, splx_data      = init_MAGEMin(db);
julia> test        = 0;
julia> sys_in      = "mol"     #default is mol, if wt is provided conversion will be done internally (MAGEMin works on mol basis)
julia> gv          = use_predefined_bulk_rock(gv, test, db)
julia> P           = 8.0;
julia> T           = 800.0;
julia> gv.verbose  = -1;        # switch off any verbose
julia> out         = point_wise_minimization(P,T, gv, z_b, DB, splx_data, sys_in)
Pressure          : 8.0      [kbar]
Temperature       : 800.0    [Celsius]
     Stable phase | Fraction (mol fraction)
              opx   0.24229
               ol   0.58808
              cpx   0.14165
              spn   0.02798
     Stable phase | Fraction (wt fraction)
              opx   0.23908
               ol   0.58673
              cpx   0.14583
              spn   0.02846
Gibbs free energy : -797.749183  (26 iterations; 94.95 ms)
Oxygen fugacity          : 9.645393319147175e-12
```

# Example 2
And here a case in which you specify your own bulk rock composition.
We convert that in the correct format, using the `convertBulk4MAGEMin` function.
```julia
julia> db          = "ig"  # database: ig, igneous (Holland et al., 2018); mp, metapelite (White et al 2014b)
julia> gv, z_b, DB, splx_data      = init_MAGEMin(db);
julia> bulk_in_ox = ["SiO2"; "Al2O3"; "CaO"; "MgO"; "FeO"; "Fe2O3"; "K2O"; "Na2O"; "TiO2"; "Cr2O3"; "H2O"];
julia> bulk_in    = [48.43; 15.19; 11.57; 10.13; 6.65; 1.64; 0.59; 1.87; 0.68; 0.0; 3.0];
julia> sys_in     = "wt"
julia> gv         = define_bulk_rock(gv, bulk_in, bulk_in_ox, sys_in, db);
julia> P,T         = 10.0, 1100.0;
julia> gv.verbose  = -1;        # switch off any verbose
julia> out         = point_wise_minimization(P,T, gv, z_b, DB, splx_data, sys_in)
Pressure          : 10.0      [kbar]
Temperature       : 1100.0    [Celsius]
     Stable phase | Fraction (mol fraction)
             pl4T   0.01114
              liq   0.74789
              cpx   0.21862
              opx   0.02154
     Stable phase | Fraction (wt fraction)
             pl4T   0.01168
              liq   0.72576
              cpx   0.23872
              opx   0.02277
Gibbs free energy : -907.27887  (47 iterations; 187.11 ms)
Oxygen fugacity          : 0.02411835177808492
julia> finalize_MAGEMin(gv,DB)
```

"""
function point_wise_minimization(P::Float64,T::Float64, gv, z_b, DB, splx_data; buffer_n = 0.0, scp = 0, rm_list = nothing, W = nothing)
    gv.buffer_n     =   buffer_n;
    input_data      =   LibMAGEMin.io_data();           # zero (not used actually)
    z_b.T           =   T + 273.15;                    # in K

    if P < 0.001
        P = 0.001
    end
    
    z_b.P           =   P
    gv.numPoint     =   1; 							    # the number of the current point */

    # Perform the point-wise minimization after resetting variables
    gv      = LibMAGEMin.reset_gv(gv,z_b, DB.PP_ref_db, DB.SS_ref_db)
    z_b     = LibMAGEMin.reset_z_b_bulk(	gv,	z_b	   )

    LibMAGEMin.reset_simplex_A(pointer_from_objref(splx_data), z_b, gv)
    LibMAGEMin.reset_simplex_B_em(pointer_from_objref(splx_data), gv)

    LibMAGEMin.reset_cp(gv,z_b, DB.cp)
    LibMAGEMin.reset_SS(gv,z_b, DB.SS_ref_db)
    LibMAGEMin.reset_sp(gv, DB.sp)

    gv      = LibMAGEMin.ComputeG0_point(gv.EM_database, z_b, gv, DB.PP_ref_db,DB.SS_ref_db);


    if ~isnothing(rm_list)
        SS_ref_db   = unsafe_wrap(Vector{LibMAGEMin.SS_ref},DB.SS_ref_db,gv.len_ss);

        for i in eachindex(rm_list)  
            flags = zeros(Int32,5);
            unsafe_copyto!(SS_ref_db[rm_list[i]].ss_flags,pointer(flags), 5)
        end
    end


    # here we can over-ride default W's
    if ~isnothing(W)
        if gv.EM_database  == W.database    # check if the database fit
        else
            print(" Wrong database number, please make sure the custom Ws are linked to the right database\n")
        end
    end


    # computing minimization
    time = @elapsed  gv      = LibMAGEMin.ComputeEquilibrium_Point(gv.EM_database, input_data, z_b, gv, pointer_from_objref(splx_data),	DB.PP_ref_db, DB.SS_ref_db, DB.cp);

    # Postprocessing (NOTE: we should switch off printing if gv.verbose=0)
    gv = LibMAGEMin.ComputePostProcessing(z_b, gv, DB.PP_ref_db, DB.SS_ref_db, DB.cp)

    # Fill structure
    LibMAGEMin.fill_output_struct(gv, pointer_from_objref(splx_data),	z_b, DB.PP_ref_db, DB.SS_ref_db, DB.cp, DB.sp );

    # Print output to screen
    LibMAGEMin.PrintOutput(gv, 0, 1, DB, time, z_b);

    # Transform results to a more convenient julia struct
    out = deepcopy(create_gmin_struct(DB, gv, time));

    if (scp == 1)
        mSS_vec     = deepcopy(out.mSS_vec)
        dT          = 2.0;
        W           = point_wise_minimization_with_guess(mSS_vec, P, T-dT, gv, z_b, DB, splx_data)
        E           = point_wise_minimization_with_guess(mSS_vec, P, T+dT, gv, z_b, DB, splx_data)
        hcp         = -(T+273.15)*(E + W - 2.0*out.G_system)/(dT*dT);
        s_cp        = hcp/out.M_sys*1e6;
        out.s_cp   .= s_cp

        # print("E: $E W: $W G: $(out.G_system)\n")
    end

    # LibMAGEMin.FreeDatabases(gv, DB, z_b);
    return out
end

"""
    out = point_wise_minimization(P::Number,T::Number, data::MAGEMin_Data)

Performs a point-wise optimization for a given pressure `P` and temperature `T` for the data specified in the MAGEMin database `MAGEMin_Data` (where also compoition is specified)
"""
point_wise_minimization(P::Number,T::Number, gv, z_b, DB, splx_data; buffer_n::Float64 = 0.0, scp::Int64 = 0, rm_list::Union{Nothing, Vector{Int64}} = nothing, W::Union{Nothing, W_Data} = nothing) = point_wise_minimization(Float64(P),Float64(T), gv, z_b, DB, splx_data; buffer_n, scp, rm_list, W)

point_wise_minimization(P::Number,T::Number, gv::LibMAGEMin.global_variables, z_b::LibMAGEMin.bulk_infos, DB::LibMAGEMin.Database, splx_data::LibMAGEMin.simplex_datas, sys_in::String; buffer_n::Float64 = 0.0, scp::Int64 = 0, rm_list::Union{Nothing, Vector{Int64}} = nothing, W::Union{Nothing, W_Data} = nothing) = point_wise_minimization(P,T, gv, z_b, DB, splx_data; buffer_n, scp, rm_list, W)

point_wise_minimization(P::Number,T::Number, data::MAGEMin_Data; buffer_n::Float64 = 0.0, scp::Int64 = 0, rm_list::Union{Nothing, Vector{Int64}} = nothing, W::Union{Nothing, W_Data} = nothing) = point_wise_minimization(P,T, data.gv[1], data.z_b[1], data.DB[1], data.splx_data[1]; buffer_n, scp, rm_list, W)


"""
    Function to provide single point minimization and give access to G0 and W's (Margules) parameters
    The objective here is to be able to use MAGEMin for thermodynamic database inversion/calibration
"""
function pwm_init(P::Float64,T::Float64, gv, z_b, DB, splx_data)

    # input_data      =   LibMAGEMin.io_data();                           # zero (not used actually)

    z_b.T           =   T + 273.15    # in K
    z_b.P           =   P

    gv.numPoint     = 1; 							    # the number of the current point */

    # Perform the point-wise minimization after resetting variables
    gv      = LibMAGEMin.reset_gv(gv,z_b, DB.PP_ref_db, DB.SS_ref_db)
    z_b     = LibMAGEMin.reset_z_b_bulk(	gv,	z_b	   )

    LibMAGEMin.reset_simplex_A(pointer_from_objref(splx_data), z_b, gv)
    LibMAGEMin.reset_simplex_B_em(pointer_from_objref(splx_data), gv)

    LibMAGEMin.reset_cp(gv,z_b, DB.cp)
    LibMAGEMin.reset_SS(gv,z_b, DB.SS_ref_db)
    LibMAGEMin.reset_sp(gv, DB.sp)

    gv      = LibMAGEMin.ComputeG0_point(gv.EM_database, z_b, gv, DB.PP_ref_db,DB.SS_ref_db);

    return (gv, z_b, DB, splx_data)
end

pwm_init(P::Number,T::Number, gv, z_b, DB, splx_data) = pwm_init(Float64(P),Float64(T), gv, z_b, DB, splx_data)


function pwm_run(gv, z_b, DB, splx_data)
    input_data      =   LibMAGEMin.io_data();                           # zero (not used actually)

    time = @elapsed  gv      = LibMAGEMin.ComputeEquilibrium_Point(gv.EM_database, input_data, z_b, gv, pointer_from_objref(splx_data),	DB.PP_ref_db,DB.SS_ref_db,DB.cp);

    # Postprocessing (NOTE: we should switch off printing if gv.verbose=0)
    gv = LibMAGEMin.ComputePostProcessing(z_b, gv, DB.PP_ref_db, DB.SS_ref_db, DB.cp)

    # Fill structure
    LibMAGEMin.fill_output_struct(gv, pointer_from_objref(splx_data), z_b, DB.PP_ref_db,DB.SS_ref_db, DB.cp, DB.sp );

    # Print output to screen
    LibMAGEMin.PrintOutput(gv, 0, 1, DB, time, z_b);

    # Transform results to a more convenient julia struct
    out = create_gmin_struct(DB, gv, time);

    # LibMAGEMin.FreeDatabases(gv, DB, z_b);

    return out
end

# pwm_run(gv, z_b, DB, splx_data) = pwm_run( gv, z_b, DB, splx_data)



"""
    structure that holds the result of the pointwise minisation

"""
struct gmin_struct{T,I}
    MAGEMin_ver :: String
    dataset     :: String
    G_system    :: T             # G of system
    Gamma       :: Vector{T}        # Gamma
    P_kbar      :: T               # Pressure in kbar
    T_C         :: T                  # Temperature in Celsius
    X           :: Vector{T}
    M_sys       :: T

    # bulk rock composition:
    bulk        :: Vector{T}
    bulk_M      :: Vector{T}
    bulk_S      :: Vector{T}
    bulk_F      :: Vector{T}

    bulk_wt     :: Vector{T}
    bulk_M_wt   :: Vector{T}
    bulk_S_wt   :: Vector{T}
    bulk_F_wt   :: Vector{T}

    # Fractions:
    # Solid, melt, fluid fractions
    frac_M      :: T
    frac_S      :: T
    frac_F      :: T

    frac_M_wt   :: T
    frac_S_wt   :: T
    frac_F_wt   :: T

    # Solid, melt, fluid densities
    alpha       :: T
    V           :: T
    cp          :: T
    s_cp        :: Vector{T}
    rho         :: T
    rho_M       :: T
    rho_S       :: T
    rho_F       :: T

    # Oxygen fugacity
    fO2         :: T
    dQFM        :: T

    # Activities
    aH2O        :: T
    aSiO2       :: T
    aTiO2       :: T
    aAl2O3      :: T
    aMgO        :: T
    aFeO        :: T

    # Phase fractions and type:
    n_PP        :: Int64                 # number of pure phases
    n_SS        :: Int64                 # number of solid solutions
    n_mSS       :: Int64                 # number of solid solutions

    ph_frac     :: Vector{T}            # phase fractions
    ph_frac_wt  :: Vector{T}            # phase fractions
    ph_frac_vol :: Vector{T}            # phase fractions
    ph_type     :: Vector{I}            # type of phase (SS or PP)
    ph_id       :: Vector{I}            # id of phase
    ph          :: Vector{String}       # Name of phase

    SS_vec      :: Vector{LibMAGEMin.SS_data}
    mSS_vec     :: Vector{LibMAGEMin.mSS_data}
    PP_vec      :: Vector{LibMAGEMin.PP_data}

    oxides      :: Vector{String}

    # Seismic velocity info
    Vp              :: T                # P-wave velocity
    Vs              :: T                # S-wave velocity
    Vp_S            :: T                # P-wave velocity of solid aggregate
    Vs_S            :: T                # S-wave velocity of solid aggregate
    bulkMod         :: T                # Elastic bulk modulus
    shearMod        :: T                # Elastic shear modulus
    bulkModulus_M   :: T                # Elastic bulk modulus
    bulkModulus_S   :: T                # Elastic bulk modulus
    shearModulus_S  :: T                # Elastic shear modulus

    # thermodynamic properties
    entropy         :: T          # entropy
    enthalpy        :: T         # enthalpy

    # Numerics:
    iter            :: I             # number of iterations required
    bulk_res_norm   :: T    # bulk residual norm
    time_ms         :: T          # computational time for this point
    status          :: I           # status of calculations
end

"""
    out = create_gmin_struct(gv, z_b)

This extracts the output of a pointwise MAGEMin optimization and adds it into a julia structure
"""
function create_gmin_struct(DB, gv, time)

    stb      = unsafe_load(DB.sp)

    MAGEMin_ver = unsafe_string(stb.MAGEMin_ver)
    dataset  = unsafe_string(stb.dataset)
    G_system = stb.G
    Gamma    = unsafe_wrap(Vector{Cdouble},stb.gamma,gv.len_ox)
    P_kbar   = stb.P
    T_C      = stb.T-273.15
    X        = [stb.X]
    M_sys    = stb.M_sys

    # Bulk rock info (total, melt, solid, fluid)
    bulk     = unsafe_wrap(Vector{Cdouble},stb.bulk,   gv.len_ox)
    bulk_M   = unsafe_wrap(Vector{Cdouble},stb.bulk_M, gv.len_ox)
    bulk_S   = unsafe_wrap(Vector{Cdouble},stb.bulk_S, gv.len_ox)
    bulk_F   = unsafe_wrap(Vector{Cdouble},stb.bulk_F, gv.len_ox)

    # Bulk rock info (total, melt, solid, fluid)
    bulk_wt     = unsafe_wrap(Vector{Cdouble},stb.bulk_wt,   gv.len_ox)
    bulk_M_wt   = unsafe_wrap(Vector{Cdouble},stb.bulk_M_wt, gv.len_ox)
    bulk_S_wt   = unsafe_wrap(Vector{Cdouble},stb.bulk_S_wt, gv.len_ox)
    bulk_F_wt   = unsafe_wrap(Vector{Cdouble},stb.bulk_F_wt, gv.len_ox)

    # Solid, melt, fluid fractions
    frac_M   = stb.frac_M
    frac_S   = stb.frac_S
    frac_F   = stb.frac_F

    # Solid, melt, fluid fractions
    frac_M_wt   = stb.frac_M_wt
    frac_S_wt   = stb.frac_S_wt
    frac_F_wt   = stb.frac_F_wt

    # Solid, melt, fluid densities
    alpha   = stb.alpha
    V       = stb.V
    cp      = stb.cp
    s_cp    = [stb.s_cp]
    rho     = stb.rho
    rho_M   = stb.rho_M
    rho_S   = stb.rho_S
    rho_F   = stb.rho_F

    # Oxygen fugacity
    fO2     = stb.fO2
    dQFM    = stb.dQFM

    # Activities
    aH2O    = stb.aH2O
    aSiO2   = stb.aSiO2
    aTiO2   = stb.aTiO2
    aAl2O3  = stb.aAl2O3
    aMgO    = stb.aMgO
    aFeO    = stb.aFeO

    # thermodynamic properties
    entropy = stb.entropy
    enthalpy= stb.enthalpy

    # Stable assemblage info
    n_ph     =  stb.n_ph        # total # of stable phases
    n_PP     =  stb.n_PP        # number of pure phases
    n_SS     =  stb.n_SS        # number of solid solutions
    n_mSS    =  stb.n_mSS        # number of solid solutions

    ph_frac  =  unsafe_wrap(Vector{Cdouble},stb.ph_frac,   n_ph)
    ph_frac_wt  =  unsafe_wrap(Vector{Cdouble},stb.ph_frac_wt,   n_ph)
    ph_frac_vol =  unsafe_wrap(Vector{Cdouble},stb.ph_frac_vol,   n_ph)
    ph_type  =  unsafe_wrap(Vector{Cint},   stb.ph_type,   n_ph)
    ph_id    =  unsafe_wrap(Vector{Cint},   stb.ph_id  ,   n_ph)
    ph       =  unsafe_string.(unsafe_wrap(Vector{Ptr{Int8}}, stb.ph, n_ph)) # stable phases

    # extract info about compositional variables of the solution models:
    SS_vec  = convert.(LibMAGEMin.SS_data, unsafe_wrap(Vector{LibMAGEMin.stb_SS_phase},stb.SS,n_SS))

    # extract information about metastable solution phases
    mSS_vec = convert.(LibMAGEMin.mSS_data, unsafe_wrap(Vector{LibMAGEMin.mstb_SS_phase},stb.mSS,n_mSS))

    # Info about the endmembers:
    PP_vec  = convert.(LibMAGEMin.PP_data, unsafe_wrap(Vector{LibMAGEMin.stb_PP_phase},stb.PP,n_PP))

    # Names of oxides:
    oxides   = unsafe_string.(unsafe_wrap(Vector{Ptr{Int8}}, stb.oxides, gv.len_ox))

    # Numerics
    bulk_res_norm   =  gv.BR_norm
    iter            =  gv.global_ite
    time_ms         =  time*1000.0

    # Store all in output struct
    out = gmin_struct{Float64,Int64}( MAGEMin_ver, dataset, G_system, Gamma, P_kbar, T_C, X, M_sys,
                bulk, bulk_M, bulk_S, bulk_F,
                bulk_wt, bulk_M_wt, bulk_S_wt, bulk_F_wt,
                frac_M, frac_S, frac_F,
                frac_M_wt, frac_S_wt, frac_F_wt,
                alpha, V, cp, s_cp,
                rho, rho_M, rho_S, rho_F,
                fO2, dQFM, aH2O, aSiO2, aTiO2, aAl2O3, aMgO, aFeO,
                n_PP, n_SS, n_mSS,
                ph_frac, ph_frac_wt, ph_frac_vol, ph_type, ph_id, ph,
                SS_vec,  mSS_vec, PP_vec,
                oxides,
                stb.Vp, stb.Vs, stb.Vp_S, stb.Vs_S, stb.bulkMod, stb.shearMod, stb.bulkModulus_M,  stb.bulkModulus_S, stb.shearModulus_S,
                entropy, enthalpy,
                iter, bulk_res_norm, time_ms, stb.status)

   return out
end


# Print brief info about pointwise calculation result
function show(io::IO, g::gmin_struct)
    println(io, "Pressure          : $(g.P_kbar)      [kbar]")
    println(io, "Temperature       : $(round(g.T_C,digits=4))    [Celsius]")

    println(io, "     Stable phase | Fraction (mol fraction) ")
    for i=1:length(g.ph)
        println(io, "   $(lpad(g.ph[i],14," "))   $( round(g.ph_frac[i], digits=5)) ")
    end
    println(io, "     Stable phase | Fraction (wt fraction) ")
    for i=1:length(g.ph)
        println(io, "   $(lpad(g.ph[i],14," "))   $( round(g.ph_frac_wt[i], digits=5)) ")
    end
    println(io, "     Stable phase | Fraction (vol fraction) ")
    for i=1:length(g.ph)
        println(io, "   $(lpad(g.ph[i],14," "))   $( round(g.ph_frac_vol[i], digits=5)) ")
    end
    println(io, "Gibbs free energy : $(round(g.G_system,digits=6))  ($(g.iter) iterations; $(round(g.time_ms,digits=2)) ms)")
    if g.status == 5
        println(io, "WARNING: calculation did not converge ----------------------------")
    end
    println(io, "Oxygen fugacity          : $(g.fO2)")
    println(io, "Delta QFM                : $(g.dQFM)")


end

"""
    print_info(g::gmin_struct)

Prints a more extensive overview of the simulation results
"""
function print_info(g::gmin_struct)

    println("Stable phases @ {$(round(g.P_kbar,digits=4)), $(round(g.T_C,digits=4))}  kbar/°C :")
    for i=1:length(g.ph)
        print("$(lpad(g.ph[i],6," ")) ")
    end
    print("  \n \n")

    # ==
    println("Compositional variables (solution phase):")
    for i=1:g.n_SS
        print("$(lpad(g.ph[i],15," ")) ")
        for j=1:length(g.SS_vec[i].compVariables)
            print("$(lpad(round(g.SS_vec[i].compVariables[j],digits=5),8," ")) ")
        end
        print("\n")
    end
    print("\n")
    # ==

    # ==
    println("End-members fraction [mol% 1 atom basis](solution phase):")
    for i=1:g.n_SS

        print("                ")
        for j=1:length(g.SS_vec[i].emNames)
            print("$(lpad(g.SS_vec[i].emNames[j],8," ")) ")
        end
        print("\n")
        print("$(lpad(g.ph[i],15," ")) ")
        for j=1:length(g.SS_vec[i].emFrac)
            print("$(lpad(round(g.SS_vec[i].emFrac[j],digits=5),8," ")) ")
        end
        print("\n")
    end
    print("\n")
    # ==
    println("End-members fraction [wt%](solution phase):")
    for i=1:g.n_SS

        print("                ")
        for j=1:length(g.SS_vec[i].emNames)
            print("$(lpad(g.SS_vec[i].emNames[j],8," ")) ")
        end
        print("\n")
        print("$(lpad(g.ph[i],15," ")) ")
        for j=1:length(g.SS_vec[i].emFrac_wt)
            print("$(lpad(round(g.SS_vec[i].emFrac_wt[j],digits=5),8," ")) ")
        end
        print("\n")
    end
    print("\n")
    # ==
      # ==
      println("Oxide compositions [mol% 1 atom basis] (normalized):")
      print("                ")
      for i=1:length(g.oxides)
          print("$(lpad(g.oxides[i],8," ")) ")
      end
      print("\n")

      print("$(lpad("SYS",15)) ")
      for i=1:length(g.oxides)
          print("$(lpad(round(g.bulk[i],digits=5),8," ")) ")
      end
      print("\n")

      for i=1:g.n_SS
          print("$(lpad(g.ph[i],15," ")) ")
          for j=1:length(g.oxides)
              print("$(lpad(round(g.SS_vec[i].Comp[j],digits=5),8," ")) ")
          end
          print("\n")
      end

      for i=1:g.n_PP
        print("$(lpad(g.ph[i],15," ")) ")
        for j=1:length(g.oxides)
            print("$(lpad(round(g.PP_vec[i].Comp[j],digits=5),8," ")) ")
        end
        print("\n")
      end
      print("\n")

      # ==
    # ==

    # ==
      # ==
      println("Oxide compositions [wt%] (normalized):")
      print("                ")
      for i=1:length(g.oxides)
          print("$(lpad(g.oxides[i],8," ")) ")
      end
      print("\n")

      print("$(lpad("SYS",15)) ")
      for i=1:length(g.oxides)
          print("$(lpad(round(g.bulk_wt[i],digits=5),8," ")) ")
      end
      print("\n")

      for i=1:g.n_SS
        print("$(lpad(g.ph[i],15," ")) ")
        for j=1:length(g.oxides)
            print("$(lpad(round(g.SS_vec[i].Comp_wt[j],digits=5),8," ")) ")
        end
        print("\n")
    end

      print("\n")
      for i=1:g.n_PP
        print("$(lpad(g.ph[i],15," ")) ")
        for j=1:length(g.oxides)
            print("$(lpad(round(g.PP_vec[i].Comp_wt[j],digits=5),8," ")) ")
        end
        print("\n")
      end
      print("\n")

      # ==
    # ==



    println("Stable mineral assemblage:")
    println("          phase  mode[mol1at] mode[wt]        f           G        V       Cp  rho[kg/m3]  Thermal_Exp Entropy[J/K] Enthalpy[J] BulkMod[GPa] ShearMod[GPa]   Vp[km/s]   Vs[km/s]")
    for i=1:g.n_SS
        print("$(lpad(g.ph[i],15," ")) ")
        print("$(lpad(round(g.ph_frac[i],digits=5),13," ")) ")
        print("$(lpad(round(g.ph_frac_wt[i],digits=5),8," ")) ")
        print("$(lpad(round(g.ph_frac_vol[i],digits=5),8," ")) ")
        print("$(lpad(round(g.SS_vec[i].f,digits=5),8," ")) ")
        print("$(lpad(round(g.SS_vec[i].G,digits=5),8," ")) ")
        print("$(lpad(round(g.SS_vec[i].V,digits=5),8," ")) ")
        print("$(lpad(round(g.SS_vec[i].cp,digits=5),8," ")) ")
        print("$(lpad(round(g.SS_vec[i].rho,digits=5),11," "))   ")
        print("$(lpad(round(g.SS_vec[i].alpha,digits=5),8," "))   ")
        print("$(lpad(round(g.SS_vec[i].entropy,digits=5),8," "))   ")
        print("$(lpad(round(g.SS_vec[i].enthalpy,digits=5),8," "))   ")
        print("$(lpad(round(g.SS_vec[i].bulkMod,digits=5),12," ")) ")
        print("$(lpad(round(g.SS_vec[i].shearMod,digits=5),13," ")) ")
        print("$(lpad(round(g.SS_vec[i].Vp,digits=5),10," ")) ")
        print("$(lpad(round(g.SS_vec[i].Vs,digits=5),10," ")) ")

        print("\n")
    end
    for i=1:g.n_PP
        print("$(lpad(g.ph[i],15," ")) ")
        print("$(lpad(round(g.ph_frac[i],digits=5),13," ")) ")
        print("$(lpad(round(g.ph_frac_wt[i],digits=5),8," ")) ")
        print("$(lpad(round(g.ph_frac_vol[i],digits=5),8," ")) ")
        print("$(lpad(round(g.PP_vec[i].f,digits=5),8," ")) ")
        print("$(lpad(round(g.PP_vec[i].G,digits=5),8," ")) ")
        print("$(lpad(round(g.PP_vec[i].V,digits=5),8," ")) ")
        print("$(lpad(round(g.PP_vec[i].cp,digits=5),8," ")) ")
        print("$(lpad(round(g.PP_vec[i].rho,digits=5),11," "))   ")
        print("$(lpad(round(g.PP_vec[i].alpha,digits=5),8," "))   ")
        print("$(lpad(round(g.PP_vec[i].entropy,digits=5),8," "))   ")
        print("$(lpad(round(g.PP_vec[i].enthalpy,digits=5),8," "))   ")
        print("$(lpad(round(g.PP_vec[i].bulkMod,digits=5),12," ")) ")
        print("$(lpad(round(g.PP_vec[i].shearMod,digits=5),13," ")) ")
        print("$(lpad(round(g.PP_vec[i].Vp,digits=5),10," ")) ")
        print("$(lpad(round(g.PP_vec[i].Vs,digits=5),10," ")) ")

        print("\n")
    end

    print("$(lpad("SYS",15," ")) ")
    print("$(lpad(round(sum(g.ph_frac),digits=5),13," ")) ")
    print("$(lpad(round(sum(g.ph_frac_wt),digits=5),8," ")) ")
    print("$(lpad(round(sum(g.ph_frac_vol),digits=5),8," ")) ")
    print("$(lpad(round(g.G_system,digits=5),20," ")) ")
    print("$(lpad(round(g.alpha,digits=5),29," ")) ")
    print("$(lpad(round(g.cp,digits=5),29," ")) ")
    print("$(lpad(round(g.s_cp[1],digits=5),29," ")) ")
    print("$(lpad(round(g.V,digits=5),29," ")) ")
    print("$(lpad(round(g.rho,digits=5),29," ")) ")
    print("$(lpad(round(g.entropy,digits=5),21," ")) ")
    print("$(lpad(round(g.enthalpy,digits=5),13," ")) ")
    print("$(lpad(round(g.bulkMod,digits=5),13," ")) ")
    print("$(lpad(round(g.shearMod,digits=5),13," ")) ")
    print("$(lpad(round(g.Vp,digits=5),10," ")) ")
    print("$(lpad(round(g.Vs,digits=5),10," ")) ")
    print("\n")
    print("\n")
    # ==

    println("Gamma (chemical potential of oxides):")
    for i=1:length(g.oxides)
        println("  $(lpad(g.oxides[i],6," "))  $(rpad(round(g.Gamma[i],digits=5),15," ")) ")
    end
    print("\n")

    println("delta Gibbs energy (G-hyperplane distance):")
    for i=1:g.n_SS
        println("  $(lpad(g.ph[i],6," "))  $(rpad(g.SS_vec[i].deltaG,15," ")) ")
    end
    for i=1:g.n_PP
        println("  $(lpad(g.ph[i],6," "))  $(rpad(g.PP_vec[i].deltaG,15," ")) ")
    end

    print("\n")

    println("mass residual :  $(g.bulk_res_norm)")
    println("# iterations  :  $(g.iter)")
    println("status        :  $(g.status)")

    print("\n")

end



function point_wise_minimization_with_guess(mSS_vec, P, T, gv, z_b, DB, splx_data)

    # initialize MAGEMin up to G0 computation included
    gv, z_b, DB, splx_data = pwm_init(P, T, gv, z_b, DB, splx_data);
    gv.verbose = -1

    ############################################################################
    # retrieve Pure Phases information
    PP_ref_db   = unsafe_wrap(Vector{LibMAGEMin.PP_ref},DB.PP_ref_db,gv.len_pp);

    # retrieve Solution Phases information
    SS_ref_db   = unsafe_wrap(Vector{LibMAGEMin.SS_ref},DB.SS_ref_db,gv.len_ss);

    # retrieve dimensions
    np          = z_b.nzEl_val
    nzEl_array  = unsafe_wrap(Vector{Cint},z_b.nzEl_array, gv.len_ox) .+ 1
    nzEl_array  = nzEl_array[1:np]

    # Declare array to be copied in splx_data
    A_jll       = zeros(np,np)
    g0_A_jll    = zeros(np)
    ph_id_A_jll = zeros(Int32,np,4)

    n_pc_ss     = zeros(gv.len_ss)

    # fill the arrays to be copied in splx_data
    for i = 1:np
        if mSS_vec[i].ph_type == "pp"
            ph_id = mSS_vec[i].ph_id+1
            g0_A_jll[i] = PP_ref_db[ph_id].gbase*PP_ref_db[ph_id].factor
            A_jll[i,:]  = mSS_vec[i].comp_Ppc[nzEl_array]

            ph_id_A_jll[i,1] = 1
            ph_id_A_jll[i,2] = ph_id-1
            ph_id_A_jll[i,3] = 0
            ph_id_A_jll[i,4] = 0
        elseif mSS_vec[i].ph_type == "ss"
            ph_id   = mSS_vec[i].ph_id+1
            ph      = mSS_vec[i].ph_name

            unsafe_copyto!(SS_ref_db[ph_id].gb_lvl,SS_ref_db[ph_id].gbase, SS_ref_db[ph_id].n_em)
            unsafe_copyto!(SS_ref_db[ph_id].iguess,pointer(mSS_vec[i].xeos_Ppc), SS_ref_db[ph_id].n_xeos)

            SS_ref_db[ph_id] = LibMAGEMin.PC_function(gv, SS_ref_db[ph_id], z_b, ph)

            g0_A_jll[i] = SS_ref_db[ph_id].df
            A_jll[i,:]  = mSS_vec[i].comp_Ppc[nzEl_array]
            ph_id_A_jll[i,1] = 3
            ph_id_A_jll[i,2] = ph_id-1
            ph_id_A_jll[i,3] = 0
            ph_id_A_jll[i,4] = n_pc_ss[ph_id]
            n_pc_ss[ph_id]  += 1
        elseif mSS_vec[i].ph_type == "ss_em"
            ph_id   = mSS_vec[i].ph_id+1
            em_id   = mSS_vec[i].em_id+1
            ape     = unsafe_wrap(Vector{Cdouble},SS_ref_db[ph_id].ape, SS_ref_db[ph_id].n_em)
            gbase   = unsafe_wrap(Vector{Cdouble},SS_ref_db[ph_id].gbase, SS_ref_db[ph_id].n_em)
            comp_ptr= unsafe_wrap(Vector{Ptr{Cdouble}},SS_ref_db[ph_id].Comp, SS_ref_db[ph_id].n_em)
            Comp    = unsafe_wrap(Vector{Cdouble},comp_ptr[em_id], gv.len_ox)
            factor 	= z_b.fbc/ape[em_id]
            ph      = mSS_vec[i].ph_name

            g0_A_jll[i] = gbase[em_id]*factor;
            A_jll[i,:]  = Comp[nzEl_array]*factor
            ph_id_A_jll[i,1] = 2
            ph_id_A_jll[i,2] = ph_id-1
            ph_id_A_jll[i,3] = 0
            ph_id_A_jll[i,4] = em_id-1
        end
    end

    # copy to the appropriate places
    ph_id_A = unsafe_wrap(Vector{Ptr{Int32}},splx_data.ph_id_A, np)

    for i=1:np
        unsafe_copyto!(ph_id_A[i],pointer(ph_id_A_jll[i,:]),4)
    end

    unsafe_copyto!(splx_data.A,pointer(vec(A_jll)),np*np)
    unsafe_copyto!(splx_data.A1,pointer(vec(A_jll)),np*np)
    unsafe_copyto!(splx_data.g0_A,pointer(g0_A_jll),np)


    # add pseudocompounds
    n_mSS = length(mSS_vec)
    for i = 1:n_mSS

        if mSS_vec[i].ph_type == "ss"
            ph          = mSS_vec[i].ph_name
            ph_id       = mSS_vec[i].ph_id+1
            n_xeos      = SS_ref_db[ph_id].n_xeos
            n_em        = SS_ref_db[ph_id].n_em

            tot_pc      = unsafe_wrap(Vector{Cint},SS_ref_db[ph_id].tot_pc, 1)
            id_pc       = unsafe_wrap(Vector{Cint},SS_ref_db[ph_id].id_pc, 1)
            info        = unsafe_wrap(Vector{Cint},SS_ref_db[ph_id].info, gv.max_n_mSS)
            factor_pc   = unsafe_wrap(Vector{Cdouble},SS_ref_db[ph_id].factor_pc, gv.max_n_mSS)
            DF_pc       = unsafe_wrap(Vector{Cdouble},SS_ref_db[ph_id].DF_pc, gv.max_n_mSS)
            G_pc        = unsafe_wrap(Vector{Cdouble},SS_ref_db[ph_id].G_pc, gv.max_n_mSS)

            m_pc        = id_pc[1]+1;
            ptr_comp_pc = unsafe_wrap(Vector{Ptr{Cdouble}},SS_ref_db[ph_id].comp_pc,gv.max_n_mSS)
            ptr_p_pc    = unsafe_wrap(Vector{Ptr{Cdouble}},SS_ref_db[ph_id].p_pc,gv.max_n_mSS)
            ptr_xeos_pc = unsafe_wrap(Vector{Ptr{Cdouble}},SS_ref_db[ph_id].xeos_pc,gv.max_n_mSS)

            unsafe_copyto!(SS_ref_db[ph_id].gb_lvl,SS_ref_db[ph_id].gbase, SS_ref_db[ph_id].n_em)
            xeos        = mSS_vec[i].xeos_Ppc

            # retrieve bounds
            bounds_ref      = zeros( n_xeos,2)
            ptr_bounds_ref  = unsafe_wrap(Vector{Ptr{Cdouble}}, SS_ref_db[ph_id].bounds_ref, n_xeos)

            for k=1:n_xeos
                bounds_ref[k,:] = unsafe_wrap(Vector{Cdouble}, ptr_bounds_ref[k], 2)
                if xeos[k] < bounds_ref[k,1]
                    xeos[k] = bounds_ref[k,1]
                elseif xeos[k] > bounds_ref[k,2]
                    xeos[k] = bounds_ref[k,2]
                end
            end

            # get solution phase information for given compositional variables
            unsafe_copyto!(SS_ref_db[ph_id].iguess,pointer(xeos), n_xeos)
            SS_ref_db[ph_id] = LibMAGEMin.PC_function(gv, SS_ref_db[ph_id], z_b, ph)

            # copy solution phase composition
            ss_comp     = unsafe_wrap(Vector{Cdouble}, SS_ref_db[ph_id].ss_comp, gv.len_ox)
            comp_pc     = unsafe_wrap(Vector{Cdouble}, ptr_comp_pc[m_pc], gv.len_ox)
            comp_pc    .= ss_comp .* SS_ref_db[ph_id].factor;

            # copy endmember fraction
            p           = unsafe_wrap(Vector{Cdouble}, SS_ref_db[ph_id].p, n_em)
            p_pc        = unsafe_wrap(Vector{Cdouble}, ptr_p_pc[m_pc], n_em)
            p_pc       .= p

            # copy compositional variables
            xeos_pc     = unsafe_wrap(Vector{Cdouble}, ptr_xeos_pc[m_pc], n_xeos)
            xeos_pc    .= xeos

            info[m_pc]      = 1;
            factor_pc[m_pc] = SS_ref_db[ph_id].factor;
            DF_pc[m_pc]     = SS_ref_db[ph_id].df;
            G_pc[m_pc]      = SS_ref_db[ph_id].df;

            tot_pc .+= 1;
            id_pc  .+= 1;
        end
    end


    gv.leveling_mode = 1
    out = deepcopy(pwm_run(gv, z_b, DB, splx_data))

    return out.G_system
end


